// (C) 2007-2025 GoodData Corporation
import { isAbortError, } from "@gooddata/sdk-backend-spi";
import { decoratedBackend } from "../decoratedBackend/index.js";
import { LRUCache } from "lru-cache";
import { DecoratedSecuritySettingsService } from "../decoratedBackend/securitySettings.js";
import { DecoratedDataView, DecoratedExecutionFactory, DecoratedExecutionResult, DecoratedPreparedExecution, } from "../decoratedBackend/execution.js";
import { DecoratedWorkspaceCatalogFactory } from "../decoratedBackend/catalog.js";
import { DecoratedElementsQuery, DecoratedElementsQueryFactory } from "../decoratedBackend/elements.js";
import stringify from "json-stable-stringify";
import compact from "lodash/compact.js";
import first from "lodash/first.js";
import flow from "lodash/flow.js";
import identity from "lodash/identity.js";
import { invariant } from "ts-invariant";
import partition from "lodash/partition.js";
import SparkMD5 from "spark-md5";
import { areObjRefsEqual, idRef, isIdentifierRef, isUriRef, uriRef, objRefToString, } from "@gooddata/sdk-model";
import { DecoratedWorkspaceAttributesService } from "../decoratedBackend/attributes.js";
import { DecoratedWorkspaceSettingsService } from "../decoratedBackend/workspaceSettings.js";
import { DecoratedWorkspaceAutomationsService, DecoratedAutomationsQuery, } from "../decoratedBackend/automations.js";
//
// Execution caching
//
class WithExecutionCaching extends DecoratedPreparedExecution {
    ctx;
    constructor(decorated, ctx, signal) {
        super(decorated, signal);
        this.ctx = ctx;
    }
    execute = async () => {
        const cacheKey = this.fingerprint();
        const cache = this.ctx.caches.execution;
        let cacheEntry = cache.get(cacheKey);
        // We need to delete the execution cache entry also if the result is cancelled,
        // to avoid 404 on result next time it's called
        // because cancel token provided from the backend was already used and we need
        // to retrieve the new one.
        const deleteExecutionCacheEntry = () => {
            cache.delete(cacheKey);
        };
        // If there is no cache entry and abort signal is provided,
        // we need cache execution only after successful resolution,
        // because running execution can be cancelled.
        if (this.signal) {
            if (!cacheEntry) {
                try {
                    const result = await super.execute().then((res) => {
                        return new WithExecutionResultCaching(res, this.createNew, this.ctx, deleteExecutionCacheEntry, this.signal);
                    });
                    cacheEntry = { result: Promise.resolve(result) };
                    cache.set(cacheKey, cacheEntry);
                }
                catch (e) {
                    deleteExecutionCacheEntry();
                    throw e;
                }
            }
        }
        else {
            if (!cacheEntry) {
                const result = super
                    .execute()
                    .then((res) => {
                    return new WithExecutionResultCaching(res, this.createNew, this.ctx, deleteExecutionCacheEntry, undefined);
                })
                    .catch((e) => {
                    deleteExecutionCacheEntry();
                    throw e;
                });
                cacheEntry = { result };
                cache.set(cacheKey, cacheEntry);
            }
        }
        return new DefinitionSanitizingExecutionResult(await cacheEntry.result, this.createNew, this.definition);
    };
    createNew = (decorated) => {
        return new WithExecutionCaching(decorated, this.ctx, this.signal);
    };
}
/**
 * This DataView decorator makes sure that definition used in it is the same as the definition in the result.
 *
 * See the usage of this class in {@link DefinitionSanitizingExecutionResult} for more.
 */
class DefinitionSanitizingDataView extends DecoratedDataView {
    constructor(decorated, result) {
        super(decorated, result);
        this.definition = result.definition;
    }
}
/**
 * This ExecutionResult decorator makes sure that definitions used throughout the result are set
 * to the definitionOverride provided. This is useful with caching because different definitions may yield
 * the same cache key, but having the proper definition in the returned execution result is critical:
 * the definitions in the result must match the one which was used to request it.
 * This however is not always the case when using cached results, so we need to ensure it explicitly.
 *
 * See the usage of this class in {@link WithExecutionCaching} for more.
 */
class DefinitionSanitizingExecutionResult extends DecoratedExecutionResult {
    constructor(decorated, wrapper, definitionOverride) {
        super(decorated, wrapper);
        this.definition = definitionOverride;
    }
    readAll = async () => {
        return this.withSanitizedDefinition(await super.readAll());
    };
    readWindow = async (offset, size) => {
        return this.withSanitizedDefinition(await super.readWindow(offset, size));
    };
    withSanitizedDefinition = (original) => {
        return new DefinitionSanitizingDataView(original, this);
    };
}
function windowKey(offset, size) {
    return `o(${offset.join(",")})_s(${size.join(",")})`;
}
class WithExecutionResultCaching extends DecoratedExecutionResult {
    ctx;
    deleteExecutionCacheEntry;
    signal;
    allData;
    allForecastConfig;
    allForecastData;
    windows;
    constructor(decorated, wrapper, ctx, deleteExecutionCacheEntry, signal) {
        super(decorated, wrapper, signal);
        this.ctx = ctx;
        this.deleteExecutionCacheEntry = deleteExecutionCacheEntry;
        this.signal = signal;
        if (cachingEnabled(this.ctx.config.maxResultWindows)) {
            this.windows = new LRUCache({ max: this.ctx.config.maxResultWindows });
        }
    }
    readAll = async () => {
        if (!this.allData && this.signal) {
            try {
                const allData = await super.readAll();
                this.allData = Promise.resolve(allData);
            }
            catch (err) {
                // If result was canceled, we need to delete also the execution cache entry,
                // to avoid 404 on result next time it's called,
                // because cancel token provided from the backend was already used and we need
                // to retrieve the new one.
                if (isAbortError(err)) {
                    this.deleteExecutionCacheEntry();
                }
                this.allData = undefined;
                throw err;
            }
        }
        else if (!this.allData) {
            this.allData = super.readAll().catch((e) => {
                this.allData = undefined;
                throw e;
            });
        }
        return this.allData;
    };
    readForecastAll = (config) => {
        // TODO: enable forecasting caching size configuration
        if (!this.allForecastData || this.allForecastConfig !== config) {
            this.allForecastConfig = config;
            this.allForecastData = super.readForecastAll(config).catch((e) => {
                this.allForecastData = undefined;
                throw e;
            });
        }
        return this.allForecastData;
    };
    readWindow = async (offset, size) => {
        if (!this.windows) {
            return super.readWindow(offset, size);
        }
        const cacheKey = windowKey(offset, size);
        let window = this.windows.get(cacheKey);
        if (this.signal && !window) {
            const result = await super.readWindow(offset, size).catch((e) => {
                if (this.windows) {
                    this.windows.delete(cacheKey);
                }
                if (isAbortError(e)) {
                    this.deleteExecutionCacheEntry();
                }
                throw e;
            });
            window = Promise.resolve(result);
            this.windows.set(cacheKey, window);
        }
        else if (!window) {
            window = super.readWindow(offset, size).catch((e) => {
                if (this.windows) {
                    this.windows.delete(cacheKey);
                }
                throw e;
            });
            this.windows.set(cacheKey, window);
        }
        return window;
    };
}
//
// Catalog caching
//
function optionsKey(options) {
    return stringify(options);
}
class WithCatalogCaching extends DecoratedWorkspaceCatalogFactory {
    ctx;
    constructor(decorated, ctx) {
        super(decorated);
        this.ctx = ctx;
    }
    load = () => {
        const cache = this.getOrCreateWorkspaceEntry(this.workspace).catalogForOptions;
        const cacheKey = optionsKey(this.options);
        let catalog = cache.get(cacheKey);
        if (!catalog) {
            catalog = super
                .load()
                .then((catalog) => {
                catalog.attributes().forEach((catalogAttribute) => {
                    catalogAttribute.displayForms.forEach((displayForm) => {
                        this.cacheAttributeByDisplayForm(displayForm.ref, catalogAttribute.attribute);
                    });
                    if (catalogAttribute.dataSet) {
                        this.cacheAttributeDataSetMeta(catalogAttribute.attribute.ref, catalogAttribute.dataSet);
                    }
                });
                return catalog;
            })
                .catch((e) => {
                cache.delete(cacheKey);
                throw e;
            });
            cache.set(cacheKey, catalog);
        }
        return catalog;
    };
    createNew = (decorated) => {
        return new WithCatalogCaching(decorated, this.ctx);
    };
    ///
    /// Caching of individual catalog objects
    ///
    cacheAttributeByDisplayForm = (displayFormRef, attribute) => {
        const cache = getOrCreateAttributeCache(this.ctx, this.workspace).attributesByDisplayForms;
        const idCacheKey = isIdentifierRef(displayFormRef) ? displayFormRef.identifier : undefined;
        let cacheItem = idCacheKey && cache.get(idCacheKey);
        if (!cacheItem) {
            cacheItem = new Promise((resolve) => resolve(attribute));
            if (idCacheKey) {
                cache.set(idCacheKey, cacheItem);
            }
        }
        return cacheItem;
    };
    cacheAttributeDataSetMeta = (attributeRef, dataSet) => {
        const cache = getOrCreateAttributeCache(this.ctx, this.workspace).dataSetsMeta;
        const idCacheKey = isIdentifierRef(attributeRef) ? attributeRef.identifier : undefined;
        let cacheItem = idCacheKey && cache.get(idCacheKey);
        if (!cacheItem) {
            cacheItem = new Promise((resolve) => resolve(dataSet));
            if (idCacheKey) {
                cache.set(idCacheKey, cacheItem);
            }
        }
        return cacheItem;
    };
    getOrCreateWorkspaceEntry = (workspace) => {
        const cache = this.ctx.caches.workspaceCatalogs;
        let cacheEntry = cache.get(workspace);
        if (!cacheEntry) {
            cacheEntry = {
                catalogForOptions: new LRUCache({
                    max: this.ctx.config.maxCatalogOptions,
                }),
            };
            cache.set(workspace, cacheEntry);
        }
        return cacheEntry;
    };
}
//
// Organization security settings caching
//
function validUrlInContextKey(url, context) {
    return `${context}_${stringify(url)}`;
}
class WithSecuritySettingsCaching extends DecoratedSecuritySettingsService {
    ctx;
    constructor(decorated, ctx) {
        super(decorated);
        this.ctx = ctx;
    }
    isUrlValid = (url, context) => {
        const cache = this.getOrCreateSecuritySettingsEntry(this.scope).valid;
        const cacheKey = validUrlInContextKey(url, context);
        let result = cache.get(cacheKey);
        if (!result) {
            result = super.isUrlValid(url, context).catch((e) => {
                cache.delete(cacheKey);
                throw e;
            });
            cache.set(cacheKey, result);
        }
        return result;
    };
    isDashboardPluginUrlValid = (url, workspace) => {
        const scope = `plugins_${workspace}`;
        const cache = this.getOrCreateSecuritySettingsEntry(scope).valid;
        let result = cache.get(url);
        if (!result) {
            result = super.isDashboardPluginUrlValid(url, workspace).catch((e) => {
                cache.delete(url);
                throw e;
            });
            cache.set(url, result);
        }
        return result;
    };
    getOrCreateSecuritySettingsEntry = (scope) => {
        const cache = this.ctx.caches.securitySettings;
        let cacheEntry = cache.get(scope);
        if (!cacheEntry) {
            cacheEntry = {
                valid: new LRUCache({
                    max: this.ctx.config.maxSecuritySettingsOrgUrls,
                    ttl: this.ctx.config.maxSecuritySettingsOrgUrlsAge,
                }),
            };
            cache.set(scope, cacheEntry);
        }
        return cacheEntry;
    };
}
class WithWorkspaceSettingsCaching extends DecoratedWorkspaceSettingsService {
    ctx;
    workspace;
    constructor(decorated, ctx, workspace) {
        super(decorated);
        this.ctx = ctx;
        this.workspace = workspace;
    }
    getSettings = () => {
        const cache = this.getOrCreateWorkspaceEntry(this.workspace).workspaceSettings;
        const cacheKey = this.workspace;
        let workspaceSettings = cache.get(cacheKey);
        if (!workspaceSettings) {
            workspaceSettings = super.getSettings().catch((e) => {
                cache.delete(cacheKey);
                throw e;
            });
            cache.set(cacheKey, workspaceSettings);
        }
        return workspaceSettings;
    };
    getSettingsForCurrentUser() {
        const cache = this.getOrCreateWorkspaceEntry(this.workspace).userWorkspaceSettings;
        const cacheKey = this.workspace; // we assume that the current user does not change over the life span of the backend instance
        let userWorkspaceSettings = cache.get(cacheKey);
        if (!userWorkspaceSettings) {
            userWorkspaceSettings = super.getSettingsForCurrentUser().catch((e) => {
                cache.delete(cacheKey);
                throw e;
            });
            cache.set(cacheKey, userWorkspaceSettings);
        }
        return userWorkspaceSettings;
    }
    async setAlertDefault(value) {
        return super.setAlertDefault(value);
    }
    async setLocale(locale) {
        return super.setLocale(locale);
    }
    async setSeparators(separators) {
        return super.setSeparators(separators);
    }
    async setTimezone(timezone) {
        return super.setTimezone(timezone);
    }
    async setDateFormat(dateFormat) {
        return super.setDateFormat(dateFormat);
    }
    async setWeekStart(weekStart) {
        return super.setWeekStart(weekStart);
    }
    async setColorPalette(colorPaletteId) {
        return super.setColorPalette(colorPaletteId);
    }
    async setTheme(themeId) {
        return super.setTheme(themeId);
    }
    deleteColorPalette() {
        return super.deleteColorPalette();
    }
    deleteTheme() {
        return super.deleteTheme();
    }
    getOrCreateWorkspaceEntry = (workspace) => {
        const cache = this.ctx.caches.workspaceSettings;
        let cacheEntry = cache.get(workspace);
        if (!cacheEntry) {
            cacheEntry = {
                userWorkspaceSettings: new LRUCache({
                    max: this.ctx.config.maxWorkspaceSettings,
                }),
                workspaceSettings: new LRUCache({
                    max: this.ctx.config.maxWorkspaceSettings,
                }),
            };
            cache.set(workspace, cacheEntry);
        }
        return cacheEntry;
    };
}
//
// Attributes caching
//
function refMatchesMdObject(ref, mdObject, type) {
    return (areObjRefsEqual(ref, mdObject.ref) ||
        areObjRefsEqual(ref, idRef(mdObject.id, type)) ||
        areObjRefsEqual(ref, uriRef(mdObject.uri)));
}
const firstDefined = flow(compact, first);
function elementsCacheKey(ref, settings) {
    return new SparkMD5().append(objRefToString(ref)).append(stringify(settings)).end();
}
function getOrCreateAttributeCache(ctx, workspace) {
    const cache = ctx.caches.workspaceAttributes;
    let cacheEntry = cache.get(workspace);
    if (!cacheEntry) {
        cacheEntry = {
            dataSetsMeta: new LRUCache({
                max: ctx.config.maxAttributesPerWorkspace,
            }),
            displayForms: new LRUCache({
                max: ctx.config.maxAttributeDisplayFormsPerWorkspace,
            }),
            attributesByDisplayForms: new LRUCache({
                max: ctx.config.maxAttributesPerWorkspace,
            }),
            attributeElementResults: cachingEnabled(ctx.config.maxAttributeElementResultsPerWorkspace)
                ? new LRUCache({
                    max: ctx.config.maxAttributeElementResultsPerWorkspace,
                })
                : undefined,
        };
        cache.set(workspace, cacheEntry);
    }
    return cacheEntry;
}
class CachedElementsQuery extends DecoratedElementsQuery {
    ctx;
    workspace;
    ref;
    constructor(decorated, ctx, workspace, ref, settings = {}) {
        super(decorated, settings);
        this.ctx = ctx;
        this.workspace = workspace;
        this.ref = ref;
    }
    query = async () => {
        const canCache = !this.settings.options?.filter;
        if (!canCache) {
            return super.query();
        }
        const cache = getOrCreateAttributeCache(this.ctx, this.workspace).attributeElementResults;
        invariant(cache, "inconsistent attribute element cache config");
        const cacheKey = elementsCacheKey(this.ref, this.settings);
        const result = cache.get(cacheKey);
        //If no cache entry and we have signal, we need cache by result
        if (this.settings.signal) {
            if (!result) {
                try {
                    const res = await super.query();
                    cache.set(cacheKey, Promise.resolve(res));
                    return res;
                }
                catch (e) {
                    cache.delete(cacheKey);
                    throw e;
                }
            }
            return result;
        }
        //If no cache entry and no signal, we need cache by promise
        if (!result) {
            const promise = super.query().catch((e) => {
                cache.delete(cacheKey);
                throw e;
            });
            cache.set(cacheKey, promise);
            return promise;
        }
        return result;
    };
    createNew(decorated, settings) {
        return new CachedElementsQuery(decorated, this.ctx, this.workspace, this.ref, settings);
    }
}
class CachedElementsQueryFactory extends DecoratedElementsQueryFactory {
    ctx;
    workspace;
    constructor(decorated, ctx, workspace) {
        super(decorated);
        this.ctx = ctx;
        this.workspace = workspace;
    }
    forDisplayForm(ref) {
        const decorated = this.decorated.forDisplayForm(ref);
        return new CachedElementsQuery(decorated, this.ctx, this.workspace, ref);
    }
}
class WithAttributesCaching extends DecoratedWorkspaceAttributesService {
    ctx;
    workspace;
    constructor(decorated, ctx, workspace) {
        super(decorated);
        this.ctx = ctx;
        this.workspace = workspace;
    }
    getAttributeDisplayForm = (ref) => {
        const cache = getOrCreateAttributeCache(this.ctx, this.workspace).displayForms;
        const idCacheKey = isIdentifierRef(ref) ? ref.identifier : undefined;
        const uriCacheKey = isUriRef(ref) ? ref.uri : undefined;
        let cacheItem = firstDefined([idCacheKey, uriCacheKey].map((key) => key && cache.get(key)));
        if (!cacheItem) {
            cacheItem = super.getAttributeDisplayForm(ref).catch((e) => {
                if (idCacheKey) {
                    cache.delete(idCacheKey);
                }
                if (uriCacheKey) {
                    cache.delete(uriCacheKey);
                }
                throw e;
            });
            if (idCacheKey) {
                cache.set(idCacheKey, cacheItem);
            }
            if (uriCacheKey) {
                cache.set(uriCacheKey, cacheItem);
            }
        }
        return cacheItem;
    };
    getAttributeDisplayForms = async (refs) => {
        const cache = getOrCreateAttributeCache(this.ctx, this.workspace).displayForms;
        // grab a reference to the cache results as soon as possible in case they would get evicted while loading the ones with missing data in cache
        // then would might not be able to call cache.get again and be guaranteed to get the data
        const refsWithCacheResults = refs.map((ref) => {
            const idCacheKey = isIdentifierRef(ref) ? ref.identifier : undefined;
            const uriCacheKey = isUriRef(ref) ? ref.uri : undefined;
            const cacheHit = firstDefined([idCacheKey, uriCacheKey].map((key) => key && cache.get(key)));
            return { ref, cacheHit };
        });
        const [withCacheHits, withoutCacheHits] = partition(refsWithCacheResults, ({ cacheHit }) => !!cacheHit);
        const refsToLoad = withoutCacheHits.map((item) => item.ref);
        const [alreadyInCache, loadedFromServer] = await Promise.all([
            // await the stuff from cache, we need the data available (we cannot just return the promises)
            Promise.all(withCacheHits.map((item) => item.cacheHit)),
            // load items not in cache using the bulk operation
            this.decorated.getAttributeDisplayForms(refsToLoad),
        ]);
        // save newly loaded to cache for future reference
        loadedFromServer.forEach((loaded) => {
            const promisifiedResult = Promise.resolve(loaded);
            // save the cache item for both types of refs
            cache.set(loaded.id, promisifiedResult);
            cache.set(loaded.uri, promisifiedResult);
        });
        const loadedRefs = loadedFromServer.map((item) => item.ref);
        const outputRefs = this.ctx.capabilities.allowsInconsistentRelations
            ? skipMissingReferences(refs, refsToLoad, loadedRefs)
            : refs;
        // reconstruct the original ordering
        const candidates = [...loadedFromServer, ...alreadyInCache];
        return outputRefs.map((ref) => {
            const match = candidates.find((item) => refMatchesMdObject(ref, item, "displayForm"));
            // if this bombs, some data got lost in the process
            invariant(match);
            return match;
        });
    };
    getAttributeByDisplayForm = async (ref) => {
        const cache = getOrCreateAttributeCache(this.ctx, this.workspace).attributesByDisplayForms;
        const idCacheKey = isIdentifierRef(ref) ? ref.identifier : undefined;
        const uriCacheKey = isUriRef(ref) ? ref.uri : undefined;
        let cacheItem = firstDefined([idCacheKey, uriCacheKey].map((key) => key && cache.get(key)));
        if (!cacheItem) {
            // eslint-disable-next-line sonarjs/no-identical-functions
            cacheItem = super.getAttributeByDisplayForm(ref).catch((e) => {
                if (idCacheKey) {
                    cache.delete(idCacheKey);
                }
                if (uriCacheKey) {
                    cache.delete(uriCacheKey);
                }
                throw e;
            });
            if (idCacheKey) {
                cache.set(idCacheKey, cacheItem);
            }
            if (uriCacheKey) {
                cache.set(uriCacheKey, cacheItem);
            }
        }
        return cacheItem;
    };
    getAttributeDatasetMeta = async (ref) => {
        const cache = getOrCreateAttributeCache(this.ctx, this.workspace).dataSetsMeta;
        const idCacheKey = isIdentifierRef(ref) ? ref.identifier : undefined;
        let cacheItem = idCacheKey && cache.get(idCacheKey);
        if (!cacheItem) {
            // eslint-disable-next-line sonarjs/no-identical-functions
            cacheItem = super.getAttributeDatasetMeta(ref).catch((e) => {
                if (idCacheKey) {
                    cache.delete(idCacheKey);
                }
                throw e;
            });
            if (idCacheKey) {
                cache.set(idCacheKey, cacheItem);
            }
        }
        return cacheItem;
    };
    async getAttributesWithReferences(refs) {
        const attributeByDfCache = getOrCreateAttributeCache(this.ctx, this.workspace).attributesByDisplayForms;
        const dataSetByAttributeCache = getOrCreateAttributeCache(this.ctx, this.workspace).dataSetsMeta;
        // grab a reference to the cache results as soon as possible in case they would get evicted while loading the ones with missing data in cache
        // then would might not be able to call cache.get again and be guaranteed to get the data
        const refsWithCacheResults = [];
        for (const ref of refs) {
            // First, get attribute cached by display form if available
            const attrCacheKey = isIdentifierRef(ref) ? ref.identifier : undefined;
            const attrCacheHit = attrCacheKey ? attributeByDfCache.get(attrCacheKey) : undefined;
            const attributeFromCache = attrCacheHit ? await attrCacheHit : undefined;
            // If attribute was cached, try to get dataSet by attribute from cache
            const dataSetCacheKey = attributeFromCache && isIdentifierRef(attributeFromCache.ref)
                ? attributeFromCache.ref.identifier
                : undefined;
            const dataSetCacheHit = dataSetCacheKey
                ? dataSetByAttributeCache.get(dataSetCacheKey)
                : undefined;
            const dataSetFromCache = dataSetCacheHit ? await dataSetCacheHit : undefined;
            // If both attribute and dataSet were cached, consider it as a cacheHit of attribute with references
            const cacheHit = attributeFromCache && dataSetFromCache
                ? Promise.resolve({
                    attribute: attributeFromCache,
                    dataSet: dataSetFromCache,
                })
                : undefined;
            refsWithCacheResults.push({ ref, cacheHit });
        }
        const [withCacheHits, withoutCacheHits] = partition(refsWithCacheResults, ({ cacheHit }) => !!cacheHit);
        const refsToLoad = withoutCacheHits.map((item) => item.ref);
        const [alreadyInCache, loadedFromServer] = await Promise.all([
            // await the stuff from cache, we need the data available (we cannot just return the promises)
            Promise.all(withCacheHits.map((item) => item.cacheHit)),
            // load items not in cache using the bulk operation
            refsToLoad.length > 0
                ? this.decorated.getAttributesWithReferences(refsToLoad)
                : Promise.resolve([]),
        ]);
        // save newly loaded to cache for future reference
        loadedFromServer.forEach((loaded) => {
            // Cache attribute by display form refs
            loaded.attribute.displayForms.forEach((displayForm) => {
                this.cacheAttributeByDisplayForm(displayForm.ref, loaded.attribute);
            });
            // Cache dataSet by attribute ref
            if (loaded.dataSet) {
                this.cacheAttributeDataSetMeta(loaded.attribute.ref, loaded.dataSet);
            }
        });
        const loadedRefs = loadedFromServer.flatMap((item) => item.attribute.displayForms.map((df) => df.ref));
        const outputRefs = this.ctx.capabilities.allowsInconsistentRelations
            ? skipMissingReferences(refs, refsToLoad, loadedRefs)
            : refs;
        // reconstruct the original ordering
        const candidates = [...loadedFromServer, ...alreadyInCache];
        return outputRefs.map((ref) => {
            const match = candidates.find((item) => item.attribute.displayForms.some((df) => areObjRefsEqual(ref, df.ref)));
            // if this bombs, some data got lost in the process
            invariant(match);
            return match;
        });
    }
    cacheAttributeByDisplayForm = (displayFormRef, attribute) => {
        const cache = getOrCreateAttributeCache(this.ctx, this.workspace).attributesByDisplayForms;
        const idCacheKey = isIdentifierRef(displayFormRef) ? displayFormRef.identifier : undefined;
        let cacheItem = idCacheKey && cache.get(idCacheKey);
        if (!cacheItem) {
            cacheItem = new Promise((resolve) => resolve(attribute));
            if (idCacheKey) {
                cache.set(idCacheKey, cacheItem);
            }
        }
        return cacheItem;
    };
    cacheAttributeDataSetMeta = (attributeRef, dataSet) => {
        const cache = getOrCreateAttributeCache(this.ctx, this.workspace).dataSetsMeta;
        const idCacheKey = isIdentifierRef(attributeRef) ? attributeRef.identifier : undefined;
        let cacheItem = idCacheKey && cache.get(idCacheKey);
        if (!cacheItem) {
            cacheItem = new Promise((resolve) => resolve(dataSet));
            if (idCacheKey) {
                cache.set(idCacheKey, cacheItem);
            }
        }
        return cacheItem;
    };
    elements() {
        const decorated = this.decorated.elements();
        return cachingEnabled(this.ctx.config.maxAttributeElementResultsPerWorkspace)
            ? new CachedElementsQueryFactory(decorated, this.ctx, this.workspace)
            : decorated;
    }
}
//AUTOMATIONS CACHING
function getOrCreateAutomationsCache(ctx, workspace) {
    const cache = ctx.caches.workspaceAutomations;
    let cacheEntry = cache.get(workspace);
    if (!cacheEntry) {
        cacheEntry = {
            automations: new LRUCache({
                max: ctx.config.maxAutomationsWorkspaces,
            }),
            queries: new LRUCache({
                max: ctx.config.maxAutomationsWorkspaces,
            }),
        };
        cache.set(workspace, cacheEntry);
    }
    return cacheEntry;
}
class CachedAutomationsQueryFactory extends DecoratedAutomationsQuery {
    ctx;
    workspace;
    settings = {
        size: 100,
        page: 0,
        author: null,
        recipient: null,
        user: null,
        dashboard: null,
        filter: {},
        sort: {},
        type: undefined,
        totalCount: undefined,
    };
    constructor(decorated, ctx, workspace) {
        super(decorated);
        this.ctx = ctx;
        this.workspace = workspace;
    }
    withSize(size) {
        this.settings.size = size;
        super.withSize(size);
        return this;
    }
    withPage(page) {
        this.settings.page = page;
        super.withPage(page);
        return this;
    }
    withFilter(filter) {
        this.settings.filter = { ...filter };
        this.settings.totalCount = undefined;
        super.withFilter(filter);
        return this;
    }
    withSorting(sort) {
        this.settings.sort = { sort };
        super.withSorting(sort);
        return this;
    }
    withType(type) {
        this.settings.type = type;
        super.withType(type);
        return this;
    }
    withAuthor(author) {
        this.settings.author = author;
        super.withAuthor(author);
        return this;
    }
    withRecipient(recipient) {
        this.settings.recipient = recipient;
        super.withRecipient(recipient);
        return this;
    }
    withUser(user) {
        this.settings.user = user;
        super.withUser(user);
        return this;
    }
    withDashboard(dashboard) {
        this.settings.dashboard = dashboard;
        super.withDashboard(dashboard);
        return this;
    }
    query() {
        const cache = getOrCreateAutomationsCache(this.ctx, this.workspace);
        const key = stringify(this.settings);
        const result = cache.queries.get(key);
        if (!result) {
            const promise = super.query().catch((e) => {
                cache.queries.delete(key);
                throw e;
            });
            cache.queries.set(key, promise);
            return promise;
        }
        return result;
    }
    async queryAll() {
        const firstQuery = await this.query();
        return firstQuery.all();
    }
}
class WithAutomationsCaching extends DecoratedWorkspaceAutomationsService {
    ctx;
    workspace;
    constructor(decorated, ctx, workspace) {
        super(decorated);
        this.ctx = ctx;
        this.workspace = workspace;
    }
    getAutomations(options) {
        const cache = getOrCreateAutomationsCache(this.ctx, this.workspace).automations;
        const key = stringify(options);
        const result = cache.get(key);
        if (!result) {
            const promise = super.getAutomations(options).catch((e) => {
                cache.delete(key);
                throw e;
            });
            cache.set(key, promise);
            return promise;
        }
        return result;
    }
    async createAutomation(automation, options) {
        const cache = getOrCreateAutomationsCache(this.ctx, this.workspace);
        const result = await super.createAutomation(automation, options);
        cache.automations.clear();
        cache.queries.clear();
        return result;
    }
    async updateAutomation(automation, options) {
        const cache = getOrCreateAutomationsCache(this.ctx, this.workspace);
        const result = await super.updateAutomation(automation, options);
        cache.automations.clear();
        cache.queries.clear();
        return result;
    }
    async deleteAutomation(id) {
        const cache = getOrCreateAutomationsCache(this.ctx, this.workspace);
        await super.deleteAutomation(id);
        cache.automations.clear();
        cache.queries.clear();
    }
    async unsubscribeAutomation(id) {
        const cache = getOrCreateAutomationsCache(this.ctx, this.workspace);
        await super.unsubscribeAutomation(id);
        cache.automations.clear();
        cache.queries.clear();
    }
    getAutomationsQuery() {
        return new CachedAutomationsQueryFactory(super.getAutomationsQuery(), this.ctx, this.workspace);
    }
}
//
//
//
function cachedExecutions(ctx) {
    return (original) => new DecoratedExecutionFactory(original, (execution) => new WithExecutionCaching(execution, ctx, execution.signal));
}
function cachedCatalog(ctx) {
    return (original) => new WithCatalogCaching(original, ctx);
}
function cachedSecuritySettings(ctx) {
    return (original) => new WithSecuritySettingsCaching(original, ctx);
}
function cachedWorkspaceSettings(ctx) {
    return (original, workspace) => new WithWorkspaceSettingsCaching(original, ctx, workspace);
}
function cachedAttributes(ctx) {
    return (original, workspace) => new WithAttributesCaching(original, ctx, workspace);
}
function cachedAutomations(ctx) {
    return (original, workspace) => new WithAutomationsCaching(original, ctx, workspace);
}
function cachingEnabled(desiredSize) {
    return desiredSize !== undefined && desiredSize > 0;
}
function cacheControl(ctx) {
    const control = {
        resetExecutions: () => {
            ctx.caches.execution?.clear();
        },
        resetCatalogs: () => {
            ctx.caches.workspaceCatalogs?.clear();
        },
        resetSecuritySettings: () => {
            ctx.caches.securitySettings?.clear();
        },
        resetAttributes: () => {
            ctx.caches.workspaceAttributes?.clear();
        },
        resetWorkspaceSettings: () => {
            ctx.caches.workspaceSettings?.clear();
        },
        resetAll: () => {
            control.resetExecutions();
            control.resetCatalogs();
            control.resetSecuritySettings();
            control.resetAttributes();
            control.resetWorkspaceSettings();
        },
    };
    return control;
}
function assertPositiveOrUndefined(value, valueName) {
    invariant(value === undefined || value > 0, `${valueName} to cache must be positive or undefined, got: ${value}`);
}
/**
 * These are the recommended settings for the backend caching.
 *
 * @remarks
 * For more information on what the options mean see {@link CachingConfiguration}.
 *
 * @public
 */
export const RecommendedCachingConfiguration = {
    maxExecutions: 10,
    maxResultWindows: 5,
    maxCatalogs: 1,
    maxCatalogOptions: 50,
    maxSecuritySettingsOrgs: 3,
    maxSecuritySettingsOrgUrls: 100,
    maxSecuritySettingsOrgUrlsAge: 300000, // 5 minutes
    maxAttributeWorkspaces: 1,
    maxAttributeDisplayFormsPerWorkspace: 500,
    maxAttributesPerWorkspace: 500,
    maxAttributeElementResultsPerWorkspace: 100,
    maxWorkspaceSettings: 1,
    maxAutomationsWorkspaces: 1,
};
/**
 * Adds caching layer on top of an existing analytical backend instance. It is currently possible to cache
 * results of executions and the workspace LDM catalog.
 *
 * @remarks see {@link CachingConfiguration} properties for more information.
 * @param realBackend - real backend to decorate with caching
 * @param config - caching configuration. {@link RecommendedCachingConfiguration} can be used
 * @public
 */
export function withCaching(realBackend, config) {
    assertPositiveOrUndefined(config.maxCatalogOptions, "maxCatalogOptions");
    assertPositiveOrUndefined(config.maxSecuritySettingsOrgUrls, "maxSecuritySettingsOrgUrls");
    assertPositiveOrUndefined(config.maxSecuritySettingsOrgUrlsAge, "maxSecuritySettingsOrgUrlsAge");
    const execCaching = cachingEnabled(config.maxExecutions);
    const catalogCaching = cachingEnabled(config.maxCatalogs);
    const securitySettingsCaching = cachingEnabled(config.maxSecuritySettingsOrgs);
    const attributeCaching = cachingEnabled(config.maxAttributeWorkspaces);
    const workspaceSettingsCaching = cachingEnabled(config.maxWorkspaceSettings);
    const automationsCaching = cachingEnabled(config.maxAutomationsWorkspaces);
    const ctx = {
        caches: {
            execution: execCaching ? new LRUCache({ max: config.maxExecutions }) : undefined,
            workspaceCatalogs: catalogCaching ? new LRUCache({ max: config.maxCatalogs }) : undefined,
            securitySettings: securitySettingsCaching
                ? new LRUCache({ max: config.maxSecuritySettingsOrgs })
                : undefined,
            workspaceAttributes: attributeCaching
                ? new LRUCache({ max: config.maxAttributeWorkspaces })
                : undefined,
            workspaceSettings: workspaceSettingsCaching
                ? new LRUCache({ max: config.maxWorkspaceSettings })
                : undefined,
            workspaceAutomations: automationsCaching
                ? new LRUCache({ max: config.maxAutomationsWorkspaces })
                : undefined,
        },
        config,
        capabilities: realBackend.capabilities,
    };
    const execution = execCaching ? cachedExecutions(ctx) : identity;
    const catalog = catalogCaching ? cachedCatalog(ctx) : identity;
    const securitySettings = securitySettingsCaching ? cachedSecuritySettings(ctx) : identity;
    const attributes = attributeCaching ? cachedAttributes(ctx) : identity;
    const automations = automationsCaching ? cachedAutomations(ctx) : identity;
    const workspaceSettings = workspaceSettingsCaching ? cachedWorkspaceSettings(ctx) : identity;
    if (config.onCacheReady) {
        config.onCacheReady(cacheControl(ctx));
    }
    return decoratedBackend(realBackend, {
        execution,
        catalog,
        securitySettings,
        attributes,
        workspaceSettings,
        automations,
    });
}
function skipMissingReferences(requestedRefs, refsToLoad, refsLoadedFromServer) {
    const missingRefs = refsToLoad.filter((ref) => !refsLoadedFromServer.some((loadedRef) => areObjRefsEqual(loadedRef, ref)));
    return requestedRefs.filter((inputRef) => !missingRefs.some((missingRef) => areObjRefsEqual(missingRef, inputRef)));
}
//# sourceMappingURL=index.js.map